# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from packaging import version
from dogtail.logging import debug_log

"""
Handles versioning of software packages
"""

__author__ = """
Dave Malcolm <dmalcolm@redhat.com>
"""

class Version:
    """
    Class representing a version of a software package.
    Stored internally as a list of subversions, from major to minor.
    Overloaded comparison operators ought to work sanely.
    """

    def __init__(self, versionList):
        debug_log("Version class constructor: '%s':'%s'" % \
            (type(versionList), str(versionList)))
        self.api_version = ".".join(str(x) for x in versionList) if isinstance(versionList, list) else versionList


    @classmethod
    def fromString(cls, versionString):
        debug_log("Getting version from Version.fromString deprecated, use constructor with string.")
        instance = Version(versionString)
        return instance


    def __str__(self):
        return self.api_version


    def __lt__(self, other):
        return version.parse(self.api_version) < version.parse(other.api_version)


    def __le__(self, other):
        return version.parse(self.api_version) <= version.parse(other.api_version)


    def __eq__(self, other):
        return version.parse(self.api_version) == version.parse(other.api_version)


    def __ne__(self, other):
        return version.parse(self.api_version) != version.parse(other.api_version)


    def __gt__(self, other):
        return version.parse(self.api_version) > version.parse(other.api_version)


    def __ge__(self, other):
        return version.parse(self.api_version) >= version.parse(other.api_version)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from io import StringIO

"""
Utility functions for 'dumping' trees of Node objects.
"""

__author__ = """
Zack Cerza <zcerza@redhat.com>,
Matěj Cepl <mcepl@redhat.com>
"""

spacer = u" "

def dump_str(node):
    with StringIO as in_memory_stream:
        plain(node, output=in_memory_stream)
        stream_buffer = in_memory_stream.getvalue()
        return stream_buffer


def plain(node, output=None):
    """
    Plain-text dump. The hierarchy is represented through indentation.
    """

    stream_output = None
    close_output = False

    def crawl(node, depth):
        do_dump(node, depth)

        actions_keys = list(node.actions.keys())
        actions_keys.sort()

        for action in actions_keys:
            do_dump(node.actions[action], depth + 1)

        for child in node.children:
            crawl(child, depth + 1)


    def dumpFile(item, depth):
        stream_output.write(str(spacer * depth) + str(item) + str("\n"))


    def dumpStdOut(item, depth):
        try:
            print(spacer * depth + str(item))  # py3
        except UnicodeDecodeError:
            print(spacer * depth + str(item).decode("utf8"))  # py2 fallback

    do_dump = dumpFile if output else dumpStdOut

    if output:
        try:
            if hasattr(output, "write"):
                stream_output = output
            elif isinstance(output, basestring):  # py2
                stream_output = open(output, "w")
                close_output = True
        except NameError:
            if isinstance(output, str):  # there's no basestring in py3 (no str and unicode)
                stream_output = open(output, "w")
                close_output = True

    crawl(node, 0)

    if close_output:
        stream_output.close()


def verbose(node, output=None):
    """
    Verbose output is the same as plain but with attributes like
    position, size, visible, showing.
    Keeping output parameter for compatibility purpose.
    Always dumping to stdout
    """

    def crawl(node, depth):
        do_dump(node, depth)
        for child in node.children:
            crawl(child, depth + 1)

    def dump_std_out(item, depth):
        print(spacer * depth + str(item) + "     [p:%s, s:%s, vis:%s, show:%s]" % \
            item.position, item.size, item.visible, item.showing)

    do_dump = dump_std_out
    crawl(node, 0)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from dogtail.config import config
from dogtail.i18n import TranslatableString
from dogtail.logging import debug_log
from dogtail.logging import debugLogger as logger
from gi.repository import GLib
import re
from time import sleep

"""
Predicates that can be used when searching for nodes.
"""

__author__ = """
David Malcolm <dmalcolm@redhat.com>
"""


def stringMatches(scriptName, reportedName):
    assert isinstance(scriptName, TranslatableString)
    return scriptName.matchedBy(reportedName)


def makeScriptRecursiveArgument(isRecursive, defaultValue):
    if isRecursive == defaultValue:
        return ""
    else:
        return ", recursive=%s" % isRecursive


def makeCamel(string):
    """
    Convert string to camelCaps
    """

    result = ""
    string = str(string)
    for char in ":;!@#$%^&*()-+=_~`\\/?|[]{}<>,.\t\n\r\"'":
        string = string.replace(char, " ")

    words = string.strip().split(" ")

    firstWord = True
    for word in words:
        lowercaseWord = word.lower()
        if firstWord:
            result += lowercaseWord
            firstWord = False
        else:
            result += lowercaseWord.capitalize()

    return result


class Predicate(object):
    """
    Abstract base class representing a predicate function on nodes.
    It's more than just a function in that it has data and can describe itself
    """

    def satisfiedByNode(self, node):
        """
        Pure virtual method returning a boolean if the predicate is satisfied by the node.
        """
        raise NotImplementedError


    def describeSearchResult(self, node):
        """
        Pure virtual method returning a string that describes the search result.
        """
        raise NotImplementedError


    def makeScriptMethodCall(self, isRecursive):
        """
        Method to generate a string containing a (hopefully) readable search
        method call on a node (to be used when generating Python source code in
        the event recorder).
        """
        raise NotImplementedError


    def makeScriptVariableName(self):
        """
        Method to generate a string containing a (hopefully) readable name
        for a Node instance variable that would be the result of a search on
        this predicate (to be used when generating Python source code in the
        event recorder).
        """
        raise NotImplementedError


    def __eq__(self, other):
        """
        Predicates are considered equal if they are of the same subclass and
        have the same data
        """

        if not isinstance(self, type(other)):
            return False
        else:
            return self.__dict__ == other.__dict__


class IsAnApplicationNamed(Predicate):
    """
    Search subclass that looks for an application by name
    """

    def __init__(self, appName):
        self.appName = TranslatableString(appName)
        self.debugName = self.describeSearchResult()
        self.satisfiedByNode = self._genCompareFunc()


    def _genCompareFunc(self):
        def satisfiedByNode(node):
            try:
                return node.roleName == "application" and stringMatches(self.appName, node.name)
            except GLib.GError as error:
                if re.match(r"name :[0-9]+\.[0-9]+ was not provided", str(error)):
                    logger.log("Dogtail: warning: omiting possibly broken at-spi application record")
                    return False

                try:
                    sleep(config.defaults["searchWarningThreshold"])
                    return node.roleName == "application" and stringMatches(self.appName, node.name)
                except GLib.GError:
                    logger.log("Dogtail: warning: application may be hanging")
                    return False
        return satisfiedByNode


    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return "%s application" % self.appName


    def makeScriptMethodCall(self, isRecursive):
        debug_log("makeScriptMethodCall(self, isRecursive=%s)" % str(isRecursive))
        return "application(%s)" % self.appName


    def makeScriptVariableName(self):
        debug_log("makeScriptVariableName(self)")
        return makeCamel(self.appName) + "App"


class GenericPredicate(Predicate):
    """
    SubtreePredicate subclass that takes various optional search fields
    """

    def __init__(self, name=None, roleName=None, description=None, label=None, debugName=None):
        if name:
            self.name = TranslatableString(name)
        else:
            self.name = None

        self.roleName = roleName
        self.description = description

        if label:
            self.label = TranslatableString(label)
        else:
            self.label = None

        if debugName:
            self.debugName = debugName
        else:
            if label:
                self.debugName = "labelled '%s'" % self.label
            else:
                self.debugName = "child with"

            if name:
                self.debugName = str(self.debugName) + str(" name=%s") % str(self.name)

            if roleName:
                self.debugName = str(self.debugName) + str(" roleName='%s'") % str(roleName)

            if description:
                self.debugName = str(self.debugName) + str(" description='%s'") % str(description)

        assert self.debugName
        self.satisfiedByNode = self._genCompareFunc()


    def _genCompareFunc(self):
        def satisfiedByNode(node):
            if self.label:
                if node.labeller:
                    return stringMatches(self.label, node.labeller.name)
                return False

            try:
                if self.name:
                    if not stringMatches(self.name, node.name):
                        return False

                if self.roleName:
                    if self.roleName != node.roleName:
                        return False

                if self.description:
                    if self.description != node.description:
                        return False
            except GLib.GError as error:
                if re.match(r"name :[0-9]+\.[0-9]+ was not provided", error):
                    logger.log("Dogtail: warning: omiting possibly broken at-spi application record")
                    return False

                raise error
            return True
        return satisfiedByNode


    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return self.debugName


    def makeScriptMethodCall(self, isRecursive):
        debug_log("makeScriptMethodCall(self, isRecursive=%s)" % str(isRecursive))

        if self.label:
            args = "label=%s" % self.label
        else:
            args = ""
            if self.name:
                print(self.name)
                args += " name=%s" % self.name

            if self.roleName:
                args += " roleName='%s'" % self.roleName

            if self.description:
                args += " description='%s'" % self.description
        return "child(%s%s)" % (args, makeScriptRecursiveArgument(isRecursive, True))


    def makeScriptVariableName(self):
        debug_log("makeScriptVariableName(self)")

        if self.label:
            return makeCamel(self.label) + "Node"

        if self.name:
            return makeCamel(self.name) + "Node"

        if self.roleName:
            return makeCamel(self.roleName) + "Node"

        if self.description:
            return makeCamel(self.description) + "Node"

        return


class IsNamed(Predicate):
    """
    Predicate subclass that looks simply by name
    """

    def __init__(self, name):
        self.name = TranslatableString(name)
        self.debugName = self.describeSearchResult()
        self.satisfiedByNode = self._genCompareFunc()


    def _genCompareFunc(self):
        def satisfiedByNode(node):
            return stringMatches(self.name, node.name)
        return satisfiedByNode


    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return "named %s" % self.name


    def makeScriptMethodCall(self, isRecursive):
        debug_log("makeScriptMethodCall(self, isRecursive=%s)" % str(isRecursive))
        return "child(name=%s%s)" % (self.name, makeScriptRecursiveArgument(isRecursive, True))


    def makeScriptVariableName(self):
        debug_log("makeScriptVariableName(self)")
        return makeCamel(self.name) + "Node"


class IsAWindowNamed(Predicate):
    """
    Predicate subclass that looks for a top-level window by name
    """

    def __init__(self, windowName):
        self.windowName = TranslatableString(windowName)
        self.debugName = self.describeSearchResult()
        self.satisfiedByNode = self._genCompareFunc()


    def _genCompareFunc(self):
        def satisfiedByNode(node):
            return node.roleName == "frame" and stringMatches(self.windowName, node.name)
        return satisfiedByNode

    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return "%s window" % self.windowName


    def makeScriptMethodCall(self, isRecursive):
        debug_log("makeScriptMethodCall(self, isRecursive=%s)" % str(isRecursive))
        return "window(%s%s)" % (self.windowName, makeScriptRecursiveArgument(isRecursive, False))


    def makeScriptVariableName(self):
        debug_log("makeScriptVariableName(self)")
        return makeCamel(self.windowName) + "Win"


class IsAWindow(Predicate):
    """
    Predicate subclass that looks for top-level windows
    """

    def __init__(self):
        self.satisfiedByNode = lambda node: node.roleName == "frame"


    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return "window"


class IsADialogNamed(Predicate):
    """
    Predicate subclass that looks for a top-level dialog by name
    """

    def __init__(self, dialogName):
        self.dialogName = TranslatableString(dialogName)
        self.debugName = self.describeSearchResult()
        self.satisfiedByNode = self._genCompareFunc()


    def _genCompareFunc(self):
        def satisfiedByNode(node):
            return node.roleName == "dialog" and stringMatches(self.dialogName, node.name)
        return satisfiedByNode


    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return "%s dialog" % self.dialogName


    def makeScriptMethodCall(self, isRecursive):
        debug_log("makeScriptMethodCall(self, isRecursive=%s)" % str(isRecursive))
        return "dialog(%s%s)" % (self.dialogName, makeScriptRecursiveArgument(isRecursive, False))


    def makeScriptVariableName(self):
        debug_log("makeScriptVariableName(self)")
        return makeCamel(self.dialogName) + "Dlg"


class IsLabelledBy(Predicate):
    """
    Predicate: is this node labelled by another node
    """


class IsLabelledAs(Predicate):
    """
    Predicate: is this node labelled with the text string (i.e. by another node with that as a name)
    """

    def __init__(self, labelText):
        self.labelText = TranslatableString(labelText)
        self.debugName = self.describeSearchResult()
        self.satisfiedByNode = self._genCompareFunc()


    def _genCompareFunc(self):
        def satisfiedByNode(node):
            if node.labeller:
                return stringMatches(self.labelText, node.labeller.name)
            return False
        return satisfiedByNode


    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return "labelled %s" % self.labelText


    def makeScriptMethodCall(self, isRecursive):
        debug_log("makeScriptMethodCall(self, isRecursive=%s)" % str(isRecursive))
        return "child(label=%s%s)" % (self.labelText, makeScriptRecursiveArgument(isRecursive, True))


    def makeScriptVariableName(self):
        debug_log("makeScriptVariableName(self)")
        return makeCamel(self.labelText) + "Node"


class IsAMenuNamed(Predicate):
    """
    Predicate subclass that looks for a menu by name
    """

    def __init__(self, menuName):
        self.menuName = TranslatableString(menuName)
        self.debugName = self.describeSearchResult()
        self.satisfiedByNode = lambda node: node.roleName == "menu" and stringMatches(self.menuName, node.name)


    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return "%s menu" % (self.menuName)


    def makeScriptMethodCall(self, isRecursive):
        debug_log("makeScriptMethodCall(self, isRecursive=%s)" % str(isRecursive))
        return "menu(%s%s)" % (self.menuName, makeScriptRecursiveArgument(isRecursive, True))


    def makeScriptVariableName(self):
        debug_log("makeScriptVariableName(self)")
        return makeCamel(self.menuName) + "Menu"


class IsAMenuItemNamed(Predicate):
    """
    Predicate subclass that looks for a menu item by name
    """

    def __init__(self, menuItemName):
        self.menuItemName = TranslatableString(menuItemName)
        self.debugName = self.describeSearchResult()
        self.satisfiedByNode = lambda node: \
            node.roleName.endswith("menu item") and stringMatches(self.menuItemName, node.name)


    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return "%s menuitem" % (self.menuItemName)


    def makeScriptMethodCall(self, isRecursive):
        debug_log("makeScriptMethodCall(self, isRecursive=%s)" % str(isRecursive))
        return "menuItem(%s%s)" % (self.menuItemName, makeScriptRecursiveArgument(isRecursive, True))


    def makeScriptVariableName(self):
        debug_log("makeScriptVariableName(self)")
        return makeCamel(self.menuItemName) + "MenuItem"


class IsATextEntryNamed(Predicate):
    """
    Predicate subclass that looks for a text entry by name
    """

    def __init__(self, textEntryName):
        self.textEntryName = TranslatableString(textEntryName)
        self.debugName = self.describeSearchResult()
        self.satisfiedByNode = lambda node: node.roleName == "text" and stringMatches(self.textEntryName, node.name)


    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return "%s textentry" % (self.textEntryName)


    def makeScriptMethodCall(self, isRecursive):
        debug_log("makeScriptMethodCall(self, isRecursive=%s)" % str(isRecursive))
        return "textentry(%s%s)" % (self.textEntryName, makeScriptRecursiveArgument(isRecursive, True))


    def makeScriptVariableName(self):
        debug_log("makeScriptVariableName(self)")
        return makeCamel(self.textEntryName) + "Entry"


class IsAButtonNamed(Predicate):
    """
    Predicate subclass that looks for a button by name
    """

    def __init__(self, buttonName):
        self.buttonName = TranslatableString(buttonName)
        self.debugName = self.describeSearchResult()
        self.satisfiedByNode = lambda node: node.roleName == "push button" and stringMatches(self.buttonName, node.name)


    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return "%s button" % (self.buttonName)


    def makeScriptMethodCall(self, isRecursive):
        debug_log("makeScriptMethodCall(self, isRecursive=%s)" % str(isRecursive))
        return "button(%s%s)" % (self.buttonName, makeScriptRecursiveArgument(isRecursive, True))


    def makeScriptVariableName(self):
        debug_log("makeScriptVariableName(self)")
        return makeCamel(self.buttonName) + "Button"


class IsATabNamed(Predicate):
    """
    Predicate subclass that looks for a tab by name
    """

    def __init__(self, tabName):
        self.tabName = TranslatableString(tabName)
        self.debugName = self.describeSearchResult()
        self.satisfiedByNode = lambda node: node.roleName == "page tab" and stringMatches(self.tabName, node.name)


    def describeSearchResult(self):
        debug_log("describeSearchResult(self)")
        return "%s tab" % (self.tabName)


    def makeScriptMethodCall(self, isRecursive):
        debug_log("makeScriptMethodCall(self, isRecursive=%s)" % str(isRecursive))
        return "tab(%s%s)" % (self.tabName, makeScriptRecursiveArgument(isRecursive, True))


    def makeScriptVariableName(self):
        debug_log("makeScriptVariableName(self)")
        return makeCamel(self.tabName) + "Tab"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import os
import sys
import time
import traceback
from dogtail.config import config

import logging
import inspect

"""
Logging facilities
"""

__author__ = """Ed Rousseau <rousseau@redhat.com>,
Zack Cerza <zcerza@redhat.com,
David Malcolm <dmalcolm@redhat.com>"""


class TimeStamp:
    """
    Timestamp class for file logs
    Generates timestamps tempfiles and log entries
    """

    def __init__(self):
        self.now = "0"
        self.timetup = time.localtime()


    def zeroPad(self, intt, width=2):
        """
        Pads an integer 'int' with zeroes, up to width 'width'.
        Returns a string.
        It will not truncate. If you call zeroPad(100, 2), '100' will be returned.
        """

        if intt < 10 ** width:
            return ("0" * (width - len(str(intt)))) + str(intt)

        return str(intt)


    def fileStamp(self, filename, addTime=True):
        """
        Generates a filename stamp in the format of filename_YYYYMMDD-hhmmss.
        A format of filename_YYYYMMDD can be used instead by specifying addTime = False.
        Should produce rel-eng style filestamps
        """

        self.now = filename.strip() + "_"
        self.timetup = time.localtime()

        fieldCount = 3
        if addTime:
            fieldCount = fieldCount + 3
        for i in range(fieldCount):
            if i == 3:
                self.now = self.now + "-"
            self.now = self.now + self.zeroPad(self.timetup[i])
        return self.now


    def entryStamp(self):
        """
        Generates a logfile entry stamp of YYYY.MM.DD HH:MM:SS
        """

        self.timetup = time.localtime()

        for i in range(6):
            if i == 0: # year
                self.now = str(self.timetup[i])
            elif i in (1, 2): # month day
                self.now = self.now + "." + self.zeroPad(self.timetup[i])
            else: # hour minutes seconds
                if i == 3:
                    self.now = self.now + " " + self.zeroPad(self.timetup[i])
                else:
                    self.now = self.now + ":" + self.zeroPad(self.timetup[i])

        return self.now


class Logger:
    """
    Writes entries to standard out.
    """

    stamper = TimeStamp()

    def __init__(self, logName, file=False, stdOut=True):
        """
        name: the name of the log
        file: The file object to log to.
        stdOut: Whether to log to standard out.
        """

        self.logName = logName
        self.stdOut = stdOut
        self.filee = file

        scriptName = config.scriptName

        if not scriptName:  # pragma: no cover
            scriptName = "log"

        self.fileName = scriptName

        if not self.filee:
            self.filee = False
            return

        if os.path.isdir(config.logDir):
            self.findUniqueName()
        else:
            raise IOError("Log path %s does not exist or is not a directory" % config.logDir)


    def findUniqueName(self):
        """
        Generate a logfile name and check if it already exists to obtain a unique one
        """

        self.fileName = config.logDir + self.stamper.fileStamp(self.fileName) + '_' + self.logName
        i = 0

        while os.path.exists(self.fileName):
            if i == 0:
                self.fileName = self.fileName + "." + str(i)
            else:
                logsplit = self.fileName.split(".")
                logsplit[-1] = str(i)
                self.fileName = ".".join(logsplit)
            i += 1


    def createFile(self):
        """
        Try to create the file and write the header info
        """

        print("Creating logfile at %s ..." % self.fileName)
        self.filee = open(self.fileName, mode="w")
        self.filee.write("##### " + os.path.basename(self.fileName) + "\n")
        self.filee.flush()


    def log(self, message, newline=True, force=False):
        """
        Hook used for logging messages. Might eventually be a virtual
        function, but nice and simple for now.

        If force is True, log to a file irrespective of config.logDebugToFile.
        """
        

        if isinstance(self.filee, bool) and (force or config.logDebugToFile):
            self.createFile()

        if force or config.logDebugToFile:
            if newline:
                self.filee.write(message + str("\n"))
            else:
                self.filee.write(message + str(" "))
            self.filee.flush()

        if self.stdOut and config.logDebugToStdOut:
            try:
                print(message)
            except TypeError:
                print(message.decode("utf-8", "replace"))

class ResultsLogger(Logger):
    """
    Writes entries into the Dogtail log
    """

    def __init__(self, stdOut=True):
        Logger.__init__(self, "results", file=True, stdOut=stdOut)


    def log(self, entry):
        """
        Writes the log entry. Requires a 1 {key: value} pair dict for an argument or else it will
        throw an exception.
        """

        if len(entry) == 1:
            key = list(entry.keys())
            value = list(entry.values())
            key = key[0]
            value = value[0]
            entry = str(key) + ":      " + str(value)
        else:
            print("Method argument requires a 1 {key: value} dict. Supplied argument not one {key: value}")
            raise ValueError(entry)

        Logger.log(self, self.stamper.entryStamp() + "      " + entry, force=True)

debugLogger = Logger("debug", config.logDebugToFile)


def exceptionHook(exc, value, tb):  # pragma: no cover
    tbStringList = traceback.format_exception(exc, value, tb)
    tbString = "".join(tbStringList)
    debugLogger.log(tbString)

sys.excepthook = exceptionHook


### Logging definition for debugging dogtail itself.
### Can be used in other parts, but we cannot break the api.

class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class DebugLogger(metaclass=Singleton):
    logger = None

    def __init__(self):
        logging.basicConfig(
            level=logging.INFO,
            format="%(message)s",
            handlers=[logging.StreamHandler()])

        self.logger = logging.getLogger("debug")


    @staticmethod
    def __get_call_info():
        stack = inspect.stack()

        file_name = stack[3][1].split("/")[-1]
        line_length = stack[3][2]

        return file_name, line_length


    def info(self, message, *args):
        message = "[{}:{:3}] {}".format(*self.__get_call_info(), message)
        self.logger.info(message, *args)

try:
    DEBUG_DOGTAIL = os.environ["DOGTAIL_DEBUG"] == "true"
except KeyError:
    DEBUG_DOGTAIL = False

_log = DebugLogger()
def debug_log(message):
    if DEBUG_DOGTAIL:
        _log.info(message)
